using System;
using System.Collections;
using System.Linq;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace CheesyFX
{
    public class Pose
    {
        public UIDynamicButton uiButton;
        public JSONClass pose;
        public byte[] img;
        private int _id;
        public EventTrigger onPoseEnter;
        public EventTrigger onPoseExit;
        private static Color selectedBGColor = new Color(0.48f, 0.48f, 0.48f);
        private static Color deselectedBGColor = new Color(0.76f, 0.76f, 0.76f, 0f);
        private static Color selectedColor = Color.white;
        public static Color deselectedColor = new Color(1f, 1f, 1f, PoseMe.buttonTransparency.val);
        private bool _selected;
        public string actors => "<b>Stored actors:</b> "+string.Join(", ", pose["pose"].AsObject.Keys.ToArray());
        public bool selected
        {
            get { return _selected;}
            set
            {
                _selected = value;
                if (value)
                {
                    uiButton.button.image.color = selectedColor;
                    backdropImage.color = selectedBGColor;
                }
                else
                {
                    uiButton.button.image.color = deselectedColor;
                    backdropImage.color = deselectedBGColor;
                }
            }
        }

        private static Atom containingAtom => PoseMe.atom;
        public int id
        {
            get { return _id; }
            set
            {
                if (value == _id) return;
                _id = value;
                if(uiButton != null) uiButton.label = _id.ToString();
            }
        }

        public Pose(int id)
        {
            if(PoseMe.currentPose != null)
            {
                PoseMe.currentPose.selected = false;
            }
            _id = id;
            onPoseEnter = new EventTrigger(PoseMe.singleton, "On Pose Enter");
            onPoseExit = new EventTrigger(PoseMe.singleton, "On Pose Exit");
            pose = GetPose();
            PoseMe.currentPose = this;
            PoseMe.onPoseEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseEnterTriggerButton.button.onClick.AddListener(onPoseEnter.OpenPanel);
            PoseMe.onPoseExitTriggerButton.button.onClick.AddListener(onPoseExit.OpenPanel);
            AddButton();
            backdropImage.color = selectedBGColor;
            selected = true;
            PoseMe.actorsJ.val = actors;
        }
        
        public Pose(int id, JSONClass pose)
        {
            this.pose = pose;
            _id = id;
            onPoseEnter = new EventTrigger(PoseMe.singleton, "On Pose Enter");
            onPoseExit = new EventTrigger(PoseMe.singleton, "On Pose Exit");
            PoseMe.currentPose = this;
            PoseMe.onPoseEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseEnterTriggerButton.button.onClick.AddListener(onPoseEnter.OpenPanel);
            PoseMe.onPoseExitTriggerButton.button.onClick.AddListener(onPoseExit.OpenPanel);
            foreach (JSONClass jc in pose["pose"][containingAtom.name]["storables"].AsArray)
            {
                if (jc["id"].Value != "control") continue;
                storedWorldToLocalMatrix = Matrix4x4.TRS(jc["position"].AsObject.ToV3(), Quaternion.Euler(jc["rotation"].AsObject.ToV3()), Vector3.one).inverse;
                break;
            }
            onPoseEnter.RestoreFromJSON(pose, PoseMe.singleton.subScenePrefix, false, true);
            onPoseExit.RestoreFromJSON(pose, PoseMe.singleton.subScenePrefix, false, true);
            if(pose.HasKey("img")) img = Convert.FromBase64String(pose["img"]);
            AddButton();
            SetButtonImage();
        }

        public void Update()
        {
            pose = GetPose();
            PoseMe.actorsJ.val = actors;
        }

        public void UpdateTriggers()
        {
            pose[onPoseEnter.Name] = onPoseEnter.GetJSON(PoseMe.singleton.subScenePrefix);
            pose[onPoseExit.Name] = onPoseExit.GetJSON(PoseMe.singleton.subScenePrefix);
        }
        
        private JSONClass GetPose()
        {
            storedWorldToLocalMatrix = containingAtom.mainController.transform.worldToLocalMatrix;
            // Matrix4x4.TRS(containingAtom.mainController.transform.position, containingAtom.mainController.transform.rotation, Vector3.one).inverse.Print();
            PoseMe.dildos.Clear();
            var jc = new JSONClass();
            jc["SceneType"] = PoseMe.sceneType;
            // jc["path"] = path.val;
            jc["id"] = id.ToString();
            var jcPose = jc["pose"] = new JSONClass();
            jcPose[containingAtom.uid] = GetPersonPose(containingAtom);
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom != containingAtom && atom.type == "Person")
                {
                    jcPose[atom.uid] = GetPersonPose(atom);
                }
                else if (atom.type == "Dildo")
                {
                    jcPose[atom.uid] = GetDildoPose(atom);
                }
            }
            if(PoseMe.manageForces.val)
            {
                var jcForces = new JSONClass();
                var hasForces = false;
                foreach (var fuckable in FillMeUp.fuckables)
                {
                    if (!fuckable.isPenetrated) continue;
                    var jcFuckable = jcForces[fuckable.name] = new JSONClass();
                    fuckable.StoreForces(jcFuckable.AsObject);
                    hasForces = true;
                }

                if (hasForces) jc["Forces"] = jcForces;
            }
            jc[onPoseEnter.Name] = onPoseEnter.GetJSON(PoseMe.singleton.subScenePrefix);
            jc[onPoseExit.Name] = onPoseExit.GetJSON(PoseMe.singleton.subScenePrefix);
            TakeScreenshot();
            return jc;
        }
        
        private JSONClass GetPersonPose(Atom atom)
        {
            JSONClass poseJSON = null;
            var storables = atom.GetStorableIDs().Select(atom.GetStorableByID)
                .Where(t => !t.exclude && t.gameObject.activeInHierarchy)
                .Where(PoseMe.ShouldStorableBeIncluded);
            var storablesJSON = new JSONArray();
            foreach (var storable in storables)
            {
                var jc = storable.GetJSON(true, false, storable.name != "geometry");
                if (atom != containingAtom && jc["id"].Value == "control")
                {
                    jc["position"] = storedWorldToLocalMatrix.MultiplyPoint(atom.mainController.transform.position).ToJC();
                    jc["rotation"] = (storedWorldToLocalMatrix.rotation * atom.mainController.transform.rotation).eulerAngles.ToJC();
                }
                storablesJSON.Add(jc);
            }

            poseJSON = new JSONClass
            {
                ["setUnlistedParamsToDefault"] = { AsBool = true },
                ["V2"] = { AsBool = true },
                ["storables"] = storablesJSON
            };
            return poseJSON;
        }

        private JSONClass GetDildoPose(Atom atom)
        {
            JSONClass jc = new JSONClass();
            jc["control"] = new JSONClass
            {
                ["pos"] = storedWorldToLocalMatrix.MultiplyPoint(atom.mainController.transform.position).ToJA(),
                ["rot"] = (storedWorldToLocalMatrix.rotation * atom.mainController.transform.rotation).eulerAngles.ToJA()
            };
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>())
            {
                jc[rb.name] = new JSONClass
                {
                    ["pos"] = storedWorldToLocalMatrix.MultiplyPoint(rb.position).ToJA(),
                    ["rot"] = (storedWorldToLocalMatrix.rotation * rb.rotation).eulerAngles.ToJA()
                };
            }
            return jc;
        }
        
        private Matrix4x4 storedWorldToLocalMatrix;
        private Matrix4x4 currentLocalToWorldMatrix;
        
        public void Apply()
        {
            PoseMe.actorsJ.val = actors;
            if (PoseMe.currentPose != null && PoseMe.currentPose != this)
            {
                PoseMe.currentPose.onPoseExit.Trigger();
                PoseMe.currentPose.selected = false;
            }
            onPoseEnter.Trigger();
            PoseMe.currentPose = this;
            PoseMe.onPoseEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseEnterTriggerButton.button.onClick.AddListener(onPoseEnter.OpenPanel);
            PoseMe.onPoseExitTriggerButton.button.onClick.AddListener(onPoseExit.OpenPanel);
            PoseMe.poseChooser.valNoCallback = PoseMe.poses.IndexOf(this).ToString();
            PoseMe.applyPosePost.Stop();
            PoseMe.softPhysicsStates.Clear();
            for (int i = 0; i < FillMeUp.persons.Count; i++)
            {
                PoseMe.softPhysicsStates.Add(PoseMe.persons[i].GetSoftPhysicsState());
            }
            currentLocalToWorldMatrix = containingAtom.mainController.transform.localToWorldMatrix;
            
            ApplyPersonPose(containingAtom);
            foreach (string atomUid in pose["pose"].AsObject.Keys)
            {
                if(atomUid == PoseMe.atom.uid) continue;
                var atom = SuperController.singleton.GetAtomByUid(atomUid);
                if (atom == null)
                {
                    $"Atom {atomUid} not found".Print();
                    continue;
                }
                if(atom.type == "Person")
                {
                    ApplyPersonPose(atom);
                }
            }
            PoseMe.applyPosePost = ApplyPosePost(pose["pose"].AsObject).Start();
            SetPreviewImage();
            selected = true;
        }

        private void ApplyPersonPose(Atom atom)
        {
            try
            {
                Quaternion headCtrlRot = Quaternion.identity;
                Quaternion headRot = Quaternion.identity;
                Quaternion neckRot = Quaternion.identity;
                Quaternion chestRot = Quaternion.identity;
                FreeControllerV3 headCtrl = null;
                DAZBone head = null;
                DAZBone neck = null;
                DAZBone chest = null;
            
                if (!PoseMe.restoreHeadRotation.val)
                {
                    headCtrl = atom.freeControllers.First(x => x.name == "headControl");
                    var bones = atom.GetComponentsInChildren<DAZBone>();
                    for (int i = 0; i < bones.Length; i++)
                    {
                        var bone = bones[i];
                        if (bone.name == "head") head = bone;
                        else if (bone.name == "neck") neck = bone;
                        else if (bone.name == "chest") chest = bone;
                    }
                    headCtrlRot = headCtrl.transform.localRotation;
                    headRot = head.transform.localRotation;
                    neckRot = neck.transform.localRotation;
                    chestRot = chest.transform.localRotation;
                }
            
                PresetManagerControl pmc = atom.presetManagerControls.First(x => x.name == "PosePresets");
                var pm = pmc.GetComponent<PresetManager>();
                
                var personJc = pose["pose"][atom.uid].AsObject;
                if(atom == containingAtom)
                {
                    
                    if (!PoseMe.restoreRoot.val)
                    {
                        
                        var controlJc = personJc["storables"].Childs.First(x => x["id"].Value == "control");
                        controlJc["id"] = "_control";
                        pm.LoadPresetFromJSON(personJc);
                        controlJc["id"] = "control";
                    }
                    else pm.LoadPresetFromJSON(personJc);
                }
                else
                {
                    Matrix4x4 refMatrix;
                    if(!PoseMe.restoreRoot.val)
                    {
                        refMatrix = currentLocalToWorldMatrix;
                    }
                    else
                    {
                        refMatrix = storedWorldToLocalMatrix.inverse;
                    }
                    var controlJc = personJc["storables"].Childs.First(x => x["id"].Value == "control");
                    var storedPos = controlJc["position"].AsObject.ToV3();
                    var storedRot = controlJc["rotation"].AsObject.ToV3();
                    
                    controlJc["position"] = refMatrix.MultiplyPoint(storedPos).ToJC();
                    controlJc["rotation"] = (refMatrix.rotation * Quaternion.Euler(storedRot)).eulerAngles.ToJC();
                    pm.LoadPresetFromJSON(personJc);
                    controlJc["position"] = storedPos.ToJC();
                    controlJc["rotation"] = storedRot.ToJC();
                }

                if (!PoseMe.restoreHeadRotation.val)
                {
                    headCtrl.transform.localRotation = headCtrlRot;
                    head.transform.localRotation = headRot;
                    neck.transform.localRotation = neckRot;
                    chest.transform.localRotation = chestRot;
                }
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }
        
        private void ApplyDildoPose(Atom atom, JSONClass pose)
        {
            Matrix4x4 refMatrix;
            if(!PoseMe.restoreRoot.val)
            {
                refMatrix = currentLocalToWorldMatrix;
            }
            else
            {
                refMatrix = storedWorldToLocalMatrix.inverse;
            }

            var controlJc = pose["control"];
            var pos = controlJc["pos"].AsArray.ToV3();
            var rot = controlJc["rot"].AsArray.ToV3();
            
            atom.mainController.transform.position = refMatrix.MultiplyPoint(pos);
            atom.mainController.transform.rotation = refMatrix.rotation * Quaternion.Euler(rot);
            
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>())
            {
                rb.position = refMatrix.MultiplyPoint(pose[rb.name]["pos"].AsArray.ToV3());
                rb.rotation = refMatrix.rotation * Quaternion.Euler(pose[rb.name]["rot"].AsArray.ToV3());
            }
        }
        
        private IEnumerator ApplyPosePost(JSONClass pose)
        {
            for (int i = 0; i < PoseMe.dildos.Count; i++)
            {
                PoseMe.dildos[i].collisionEnabled = false;
            }
            yield return new WaitUntil(() =>
            {
                for (int i = 0; i < PoseMe.persons.Count; i++)
                {
                    if (PoseMe.persons[i].GetSoftPhysicsState() != PoseMe.softPhysicsStates[i]) return false;
                }
                return true;
            });
            foreach (string atomUid in pose.Keys)
            {
                var atom = SuperController.singleton.GetAtomByUid(atomUid);
                if (atom == null) continue;
                var atomPose = pose[atomUid].AsObject;
                if (atom.type == "Person")
                {
                    // JSONArray storables = atomPose["storables"].AsArray;
                    // foreach (var bone in atom.GetComponentsInChildren<DAZBone>())
                    // {
                    //     
                    //     for (int i = 0; i < storables.Count; i++)
                    //     {
                    //         JSONClass jc = storables[i].AsObject;
                    //         if(jc["id"] != bone.name) continue;
                    //         bone.Print();
                    //         bone.RestoreFromJSON(jc);
                    //     }
                    // }
                }
                else if(atom.type == "Dildo")
                {
                    ApplyDildoPose(atom, pose[atom.uid].AsObject);
                }
            }
            for (int i = 0; i < PoseMe.dildos.Count; i++)
            {
                PoseMe.dildos[i].collisionEnabled = true;
            }

            if (PoseMe.manageForces.val && pose.HasKey("Forces"))
            {
                var jcForces = pose["Forces"].AsObject;
                foreach (var fuckable in FillMeUp.fuckables)
                {
                    if(!jcForces.HasKey(fuckable.name)) continue;
                    fuckable.LoadForces(jcForces[fuckable.name].AsObject);
                }
            }
        }

        public RectTransform buttonRT;
        private Image backdropImage;
        
        private void AddButton()
        {
            buttonRT = Object.Instantiate(PoseMe.singleton.manager.configurableButtonPrefab) as RectTransform;
            Object.Destroy(buttonRT.GetComponentInChildren<UIDynamicButton>());
            Object.Destroy(buttonRT.GetComponentInChildren<Button>());
            Object.Destroy(buttonRT.GetComponentInChildren<Text>());
            backdropImage = buttonRT.GetComponent<Image>();
            backdropImage.color = deselectedBGColor;
            buttonRT.SetParent(PoseMe.canvas.transform, false);
            
            RectTransform button = Object.Instantiate(PoseMe.singleton.manager.configurableButtonPrefab) as RectTransform;
            // button.SetParent(PoseMe.canvas.transform, false);
            button.SetParent(buttonRT, false);
            button.anchorMax = new Vector2(.98f, .98f);
            button.anchorMin = new Vector2(.02f, .02f);
            button.offsetMax = new Vector2();
            button.offsetMin = new Vector2();

            uiButton = button.GetComponent<UIDynamicButton>();
            uiButton.label = _id.ToString();
            uiButton.buttonText.fontSize = 30;
            uiButton.buttonText.color = Color.black;
            uiButton.button.image.color = deselectedColor;
            uiButton.button.onClick.AddListener(Apply);
            var cb = uiButton.button.colors;
            cb.pressedColor = cb.highlightedColor = cb.normalColor;
            uiButton.button.colors = cb;
        }

        public void RemoveButton()
        {
            Object.Destroy(uiButton.transform.parent.gameObject);
        }

        public void StoreToDisk(string path)
        {
            if(img != null) FileManagerSecure.WriteAllBytes($"{path}.jpg", img);
        }

        // public void DeleteFromDisk()
        // {
        //     if (!FileManagerSecure.FileExists(path.val + ".json")) return;
        //     FileManagerSecure.DeleteFile(path.val+".json");
        //     FileManagerSecure.DeleteFile(path.val+".jpg");
        // }

        public void StoreBackup()
        {
            var backupPath = $"{PoseMe.backUpStorePath}{PoseMe.sceneType}_{DateTime.Now:yyyy-MM-dd-HHMMss}";
            if(img != null) FileManagerSecure.WriteAllBytes(backupPath+".jpg", img);
            SuperController.singleton.SaveJSON(pose, backupPath+".json");
            // FileManagerSecure.MoveFile(path.val+".json", backupName + ".json");
            // FileManagerSecure.MoveFile(path.val + ".jpg", backupName + ".jpg");
            $"PoseMe: Backup stored at {backupPath}.".Print();
        }

        public void SyncAtomNames(string oldUid, string newUid)
        {
            onPoseExit.SyncAtomNames();
            onPoseExit.SyncAtomNames();
            var atomPoses = pose["pose"];
            atomPoses[newUid] = atomPoses[oldUid];
            atomPoses.Remove(oldUid);
        }

        private void TakeScreenshot()
        {
            SuperController.singleton.DoSaveScreenshot(PoseMe.tmpPath+"1.jpg", OnScreenshotTaken);
        }

        private void OnScreenshotTaken(string imagePath)
        {
            if(!FileManagerSecure.FileExists(imagePath)) return;
            img = FileManagerSecure.ReadAllBytes(imagePath);
            FileManagerSecure.DeleteFile(imagePath);
            pose["img"] = Convert.ToBase64String(img);
            // SuperController.singleton.SaveJSON(pose, $"{path.val}_.json");
            SetButtonImage(PoseMe.showThumbnails.val);
            SetPreviewImage();
        }
        
        // public void FetchThumbnail(string path)
        // {
        //     img = FileManagerSecure.ReadAllBytes(path);
        //     FileManagerSecure.WriteAllBytes(this.path.val+".jpg", img);
        //     SetButtonImage(PoseMe.showThumbnails.val);
        // }
        
        public void SetButtonImage(bool showThumbnail = true)
        {
            if(showThumbnail && img != null)
            {
                var texture = LoadTexture(img);
                uiButton.button.image.sprite = Sprite.Create(texture,
                    new Rect(0.0f, 0.0f, texture.width, texture.height), new Vector2(0.5f, 0.5f));
                uiButton.buttonText.alignment = TextAnchor.UpperRight;
                uiButton.buttonText.color = Color.magenta;
                uiButton.buttonText.fontSize = 30;
                
            }
            else
            {
                uiButton.button.image.sprite = null;
                uiButton.buttonText.alignment = TextAnchor.MiddleCenter;
                uiButton.buttonText.color = Color.black;
                uiButton.buttonText.fontSize = 40;
            }
        }

        private void SetPreviewImage()
        {
            if(img == null) return;
            var texture = LoadTexture(img);
            PoseMe.previewImage.sprite = Sprite.Create(texture,
                new Rect(0.0f, 0.0f, texture.width, texture.height), new Vector2(0.5f, 0.5f));
        }
        
        public static Texture2D LoadTexture(string path)
        {
            var tex = new Texture2D(2, 2);
            var data = FileManagerSecure.ReadAllBytes(path);
            tex.LoadImage(data);
            
            // var pixels = tex.GetPixels();
            // for (int i = 0; i < pixels.Length; i++)
            // {
            //     var pixel = pixels[i];
            //     
            //     if (pixel.ToHSV().V < .1f)
            //     {
            //         // pixel.ToHSV().V.Print();
            //         // pixel.a = 0f;
            //         // pixels[i] = pixel;
            //         var col = Color.white;
            //         col.a = 0f;
            //         pixels[i] = col;
            //     }
            //     // pixels[i] *= Color.Lerp(color, pixels[i], maskPixels[i].a);
            // }
            // tex.SetPixels(pixels);
            tex.Apply();
            return tex;
        }

        private static Texture2D LoadTexture(byte[] image)
        {
            var tex = new Texture2D(2, 2);
            tex.LoadImage(image);
            tex.Apply();
            return tex;
        }
    }
}